#include <algorithm>
#include <map>
#include "HistogramOp.h"

const float HistogramOp::PRECISION = 0.000001f;

float HistogramOp::minGrayValue(Image &img)
{
	int imgSize = img.size();
	float val, min = img.Pix(0);

	for(int i=1; i<imgSize; i++) {
		val = img.Pix(i);
		if(val < min)
			min = val;
	}
	return(min);
}
/*****************************************************************************/
float HistogramOp::minGrayValue(vector<Image *> &imgs)
{
	int imgSize, imgNum = imgs.size();
	float val, min;

	min = imgs[0]->Pix(0);
	for(int j=0; j<imgNum; j++) {
		Image &img = *(imgs[j]);
		imgSize = img.size();
		for(int i=0; i<imgSize; i++) {
			val = img.Pix(i);
			if(val < min)
				min = val;
		}
	}
	return(min);
}
/*****************************************************************************/
float HistogramOp::maxGrayValue(Image &img)
{
	int imgSize = img.size();
	float val, max = img.Pix(0);

	for(int i=1; i<imgSize; i++) {
		val = img.Pix(i);
		if(val > max)
			max = val;
	}
	return(max);
}
/*****************************************************************************/
float HistogramOp::maxGrayValue(vector<Image *> &imgs)
{
	int imgSize, imgNum = imgs.size();
	float val, max;

	max = imgs[0]->Pix(0);
	for(int j=0; j<imgNum; j++) {
		Image &img = *(imgs[j]);
		imgSize = img.size();
		for(int i=0; i<imgSize; i++) {
			val = img.Pix(i);
			if(val > max)
				max = val;
		}
	}
	return(max);
}
/*****************************************************************************/
float HistogramOp::min(Image &img)
{
  vector<HistoRec *> histogram;
  int i,histSize;

  createHistogram(img,histogram);
  histSize = histogram.size();

  float min = histogram[0]->value;
  int times = histogram[0]->times;
  
  for(i=1; i<histSize; i++) {
    if(histogram[i]->times < times) {
      min = histogram[i]->value;
      times = histogram[i]->times;
    }
  }
	            //free allocated memory
  for(i=0; i<histSize; i++)
    delete histogram[i];

  return(min);
}
/*****************************************************************************/
float HistogramOp::min(vector<Image *> &imgs)
{
  vector<HistoRec *> histogram;
  int i,histSize;

  createHistogram(imgs,histogram);
  histSize = histogram.size();

  float min = histogram[0]->value;
  int times = histogram[0]->times;
  
  for(i=1; i<histSize; i++) {
    if(histogram[i]->times < times) {
      min = histogram[i]->value;
      times = histogram[i]->times;
    }
  }
	            //free allocated memory
  for(i=0; i<histSize; i++)
    delete histogram[i];

  return(min);
}
/*****************************************************************************/
float HistogramOp::mode(Image &img)
{
  vector<HistoRec *> histogram;
  int i,histSize;

  createHistogram(img,histogram);
  histSize = histogram.size();

  float mode = histogram[0]->value;
  int times = histogram[0]->times;
  
  for(i=1; i<histSize; i++) {
    if(histogram[i]->times > times) {
      mode = histogram[i]->value;
      times = histogram[i]->times;
    }
  }
	            //free allocated memory
  for(i=0; i<histSize; i++)
    delete histogram[i];

  return(mode);
}
/*****************************************************************************/
float HistogramOp::mode(vector<Image *> &imgs)
{
  vector<HistoRec *> histogram;
  int i,histSize;

  createHistogram(imgs,histogram);
  histSize = histogram.size();

  float mode = histogram[0]->value;
  int times = histogram[0]->times;
  
  for(i=1; i<histSize; i++) {
    if(histogram[i]->times > times) {
      mode = histogram[i]->value;
      times = histogram[i]->times;
    }
  }
	            //free allocated memory
  for(i=0; i<histSize; i++)
    delete histogram[i];

  return(mode);
}
/*****************************************************************************/
void HistogramOp::normalizeClamp(Image &img, 
																 float lowerThresholdPercentage,
																 float upperThresholdPercentage,
																 float minVal, float maxVal)
{
  vector<HistoRec *> histogram;
  int i,total,curNum;
  float lower,upper;
  
  total = img.size();
  unsigned lowerThreshold = (unsigned)(lowerThresholdPercentage*total+0.5);
  unsigned upperThreshold = (unsigned)(upperThresholdPercentage*total+0.5);
  
  createHistogram(img,histogram);
	int histSize = histogram.size();
	
  curNum = 0;
  i=0;
  while(curNum < lowerThreshold)
    curNum+=histogram[i++]->times;
  lower = i==histSize ? histogram[i-1]->value : histogram[i]->value;
  
  curNum = 0;
  i = histogram.size() - 1;
  while(curNum < upperThreshold)
    curNum+=histogram[i--]->times;
  upper = i==-1 ? histogram[i+1]->value : histogram[i]->value;
  
  if(upper > lower) {
    float scale = (maxVal-minVal)/(upper-lower);
    float val;
    for(i=0; i<total; i++) {
      val = img.Pix(i);
      if(val > upper)
				img.Pix(i) = maxVal;
      else if(val < lower)
				img.Pix(i) = minVal;
      else
				img.Pix(i) = scale*(val - lower) + minVal;
    }
  }
	//free allocated memory
  for(i=0; i<histSize; i++)
    delete histogram[i];
}
/*****************************************************************************/
void HistogramOp::normalizeClamp(vector<Image *> &imgs, 
																 float lowerThresholdPercentage,
																 float upperThresholdPercentage,
																 float minVal, float maxVal)
{
  vector<HistoRec *> histogram;
  int i, j, total, curNum, imgNum, imgSize;
  float lower,upper;
  
	imgNum = imgs.size();
  total=0;
	for(i=0; i<imgNum; i++)
		total+=imgs[i]->size();
  unsigned lowerThreshold = (unsigned)(lowerThresholdPercentage*total+0.5);
  unsigned upperThreshold = (unsigned)(upperThresholdPercentage*total+0.5);
  
  createHistogram(imgs,histogram);
	int histSize = histogram.size();
	
  curNum = 0;
  i=0;
  while(curNum < lowerThreshold)
    curNum+=histogram[i++]->times;
  lower = i==histSize ? histogram[i-1]->value : histogram[i]->value;
  
  curNum = 0;
  i = histogram.size() - 1;
  while(curNum < upperThreshold)
    curNum+=histogram[i--]->times;
  upper = i==-1 ? histogram[i+1]->value : histogram[i]->value;
  
  if(upper > lower) {
    float scale = (maxVal-minVal)/(upper-lower);
    float val;
		for(j=0; j<imgNum; j++) {      //go over all images 
			Image & img = *(imgs[j]);
			imgSize = img.size();
			for(i=0; i<imgSize; i++) {     //go over all pixels
				val = img.Pix(i);
				if(val > upper)
					img.Pix(i) = maxVal;
				else if(val < lower)
					img.Pix(i) = minVal;
				else
					img.Pix(i) = scale*(val - lower) + minVal;
			}
		}
  }
	//free allocated memory
  for(i=0; i<histSize; i++)
    delete histogram[i];
}
/*****************************************************************************/
void HistogramOp::histEqualize(Image &img, float minVal, float maxVal)
{
  vector<HistoRec *> histogram;
  float SkMin,Sk;
  float range = maxVal - minVal;
  int total = img.size();
  int i,histSize,sum;
	

                         //compute equalization mapping
  createHistogram(img,histogram);

  FloatComparator floatComp(HistogramOp::PRECISION);
  std::map<float, float, FloatComparator> *mapping = 
    new std::map< float, float, FloatComparator >(floatComp);
  SkMin = (float)(histogram[0]->times)/(float)total;
  histSize = histogram.size();
  
                  //special case where the whole image has only one value
  if((1.0 - SkMin) < HistogramOp::PRECISION)
    mapping->insert(std::map<float, float, 
		    FloatComparator>::value_type(histogram[0]->value,
						 histogram[0]->value));
  else {
    sum = 0;
    for(i=0; i<histSize; i++) {
      sum += histogram[i]->times;
      Sk = (((float)sum/(float)total) - SkMin) / (1.0-SkMin);
      mapping->insert(std::map<float, float, 
		      FloatComparator>::value_type(histogram[i]->value,
						   Sk*range+minVal));
    }
  }

     	//apply equalization to the image
  std::map<float, float, FloatComparator>::iterator it;
  for(i=0; i<total; i++) {
    it = mapping->find(img.Pix(i));
    img.Pix(i) = (*it).second;
  }
	  
	    //free allocated memory
  delete mapping;
  for(i=0; i<histSize; i++) 
    delete histogram[i];
}
/*****************************************************************************/
void HistogramOp::histEqualize(vector<Image *> &imgs, float minVal, float maxVal)
{
  vector<HistoRec *> histogram;
  float SkMin,Sk;
  float range = maxVal - minVal;
  int i, j, histSize, sum, total, imgNum, imgSize;
	
	imgNum = imgs.size();
  total=0;
	for(i=0; i<imgNum; i++)
		total+=imgs[i]->size();

          	//compute equalization mapping
  createHistogram(imgs,histogram);

  FloatComparator floatComp(HistogramOp::PRECISION);
  std::map<float, float, FloatComparator> *mapping = 
    new std::map< float, float, FloatComparator >(floatComp);
  SkMin = (float)(histogram[0]->times)/(float)total;
  histSize = histogram.size();
  
                  //special case where the whole image has only one value
  if((1.0 - SkMin) < HistogramOp::PRECISION)
    mapping->insert(std::map<float, float, 
		    FloatComparator>::value_type(histogram[0]->value,
						 histogram[0]->value));
  else {
    sum = 0;
    for(i=0; i<histSize; i++) {
      sum += histogram[i]->times;
      Sk = (((float)sum/(float)total) - SkMin) / (1.0-SkMin);
      mapping->insert(std::map<float, float, 
		      FloatComparator>::value_type(histogram[i]->value,
						   Sk*range+minVal));
    }
  }

     	//apply equalization to the image
  std::map<float, float, FloatComparator>::iterator it;

	for(j=0; j<imgNum; j++) {  //go over all images
		Image &img = *(imgs[j]);
		imgSize = img.size();
		for(i=0; i<imgSize; i++) {
			it = mapping->find(img.Pix(i));
			img.Pix(i) = (*it).second;
		}
	}	  
	    //free allocated memory
  delete mapping;
  for(i=0; i<histSize; i++) 
    delete histogram[i];
}
/*****************************************************************************/
/*
 * Create the histogram as follows:
 *     1. init empty map.
 *     2. for each gray value
 *           if found in map update the number of times it appears.
 *           else add new entry (pointer) into the map and the vector.
 *     3. sort the vector.
 *
 * Accessing existing entries through the map allows for fast update [O(logn)].
 */
void HistogramOp::createHistogram(Image &img, vector<HistoRec *> &histogram)
{
  int i,sz;
  float value;
  HistoRec* entry;

  FloatComparator floatComp(HistogramOp::PRECISION);
  HistoRecValComparator histoRVComparator(&floatComp);

  std::map<float, HistoRec*, FloatComparator> *histoEntries = 
    new std::map< float, HistoRec*, FloatComparator >(floatComp);
  std::map<float, HistoRec*, FloatComparator>::iterator it;

  histogram.clear();
  sz = img.size();
  for(i=0; i<sz; i++) {
    value = img.Pix(i);
    it = histoEntries->find(value);
    if(it == histoEntries->end()) { //new entry
      entry = new HistoRec(value);
      histogram.push_back(entry);
      histoEntries->insert(std::map<float, HistoRec*, 
			   FloatComparator>::value_type(value,entry));	                                                                           
    }
    else   //update existing entry
      (*it).second->times++;
  }
  sort(histogram.begin(),histogram.end(),histoRVComparator);
  delete histoEntries;
}
/*****************************************************************************/
/*
 * Create the histogram as follows:
 *     1. init empty map.
 *     2. for each gray value
 *           if found in map update the number of times it appears.
 *           else add new entry (pointer) into the map and the vector.
 *     3. sort the vector.
 *
 * Accessing existing entries through the map allows for fast update [O(logn)].
 */

void HistogramOp::createHistogram(vector<Image *> &imgs, vector<HistoRec *> &histogram)
{
  int i, j, sz, imgNum;
  float value;
  HistoRec* entry;

  FloatComparator floatComp(HistogramOp::PRECISION);
  HistoRecValComparator histoRVComparator(&floatComp);

  std::map<float, HistoRec*, FloatComparator> *histoEntries = 
    new std::map< float, HistoRec*, FloatComparator >(floatComp);
  std::map<float, HistoRec*, FloatComparator>::iterator it;

  histogram.clear();
	imgNum = imgs.size();
	for(j=0; j<imgNum; j++) {  //go over all images 
		Image & img = *(imgs[j]);
		sz = img.size();
		for(i=0; i<sz; i++) {   //go over all pixels 
			value = img.Pix(i);
			it = histoEntries->find(value);
			if(it == histoEntries->end()) { //new entry
				entry = new HistoRec(value);
				histogram.push_back(entry);
				histoEntries->insert(std::map<float, HistoRec*, 
					 FloatComparator>::value_type(value,entry));	                                                                           
			}
			else   //update existing entry
				(*it).second->times++;
		}
	}
  sort(histogram.begin(),histogram.end(),histoRVComparator);
  delete histoEntries;
}
/*****************************************************************************/
void HistogramOp::printHistogram(vector<HistoRec *> &histogram)
{
  for(int i=0; i<histogram.size(); i++)
    printf("%f : %d\n",histogram[i]->value,histogram[i]->times);
}
